# -*- encoding: utf-8 -*-
#
# Copyright © 2016-2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import six


class UnableToDecodeBody(Exception):
    def __init__(self, exception, body):
        self.reason = six.text_type(exception)
        self.body = body
        super(UnableToDecodeBody, self).__init__(body)

    def jsonify(self):
        return {
            "cause": "Unable to decode body",
            "reason": self.reason,
            "detail": self.body
        }
